"""
Common interface for table detectors.
"""

from os import PathLike
from typing import Callable, Iterator, List, Protocol, Tuple, Union
from playa import Document, Page, PageList

from paves.tables.table import TableObject


class Detector(Protocol):
    """Protocol for table detectors."""

    def __call__(
        self,
        pdf: Union[str, PathLike, Document, Page, PageList],
    ) -> Union[Iterator[TableObject], None]: ...

    __name__: str


DETECTORS: List[Tuple[int, Detector]] = []


def detector(*, priority: int) -> Callable[[Detector], Detector]:
    """Decorator to register detector functions with priorities."""

    def register(func: Detector) -> Detector:
        DETECTORS.append((priority, func))
        # We don't care about the inefficiency of this as there are
        # only ever going to be a few of them
        DETECTORS.sort()
        return func

    return register


def lookup(name: str) -> Union[Detector, None]:
    """Look up a detector by name."""
    for _, d in DETECTORS:
        if d.__name__ == name:
            return d
    return None


def tables_orelse(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Union[Iterator[TableObject], None]:
    """Identify tables in a PDF or one of its pages, or fail.

    This works like `tables` but forces you (if you use type checking)
    to detect the case where tables cannot be detected by any known
    method.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.

    Returns:
        An iterator over `TableObject`, or `None`, if there is no
        method available to detect tables.  This will cause a
        `TypeError` if you try to iterate over it anyway.

    """
    for _, method in DETECTORS:
        itor = method(pdf)
        if itor is not None:
            return itor
    else:
        return None


def tables(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Iterator[TableObject]:
    """Identify tables in a PDF or one of its pages.

    This will always try to use logical structure (via PLAYA-PDF)
    first to identify tables.

    Of course, that only works on tagged and accessible PDFs.  So,
    like `paves.image`, we can also use Machine Learning Models™ here,
    which involves nasty horrible dependencyses (we hates them, they
    stole the precious) like `cudnn-10-gigabytes-of-c++`.

    If you'd like to try that, then you can do so by installing the
    `transformers[torch]` package (if you don't have a GPU, try adding
    `--extra-index-url https://download.pytorch.org/whl/cpu` to pip's
    command line).

    Note: These tables cannot span multiple pages.
        Often, a table will span multiple pages.  With PDF logical
        structure, this can be represented (and sometimes is), but if
        there is no logical structure, this is not possible, since
        tables are detected from the rendered image of a page.
        Reconstructing this information is both extremely important
        and also very difficult with current models (perhaps very big
        VLMs can do it?).  Since we also want to visualize tables with
        `paves.image`, we don't return multi-page tables here.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.

    Returns:
        An iterator over `TableObject`.  If no method is available to
        detect tables, this will return an iterator over an empty
        list.  You may wish to use `tables_orelse` to ensure that
        tables can be detected.

    """
    itor = tables_orelse(pdf)
    if itor is None:
        return iter(())
    return itor



from typing import Callable, Iterator, Tuple, Union

from PIL import Image

class PdfBitmap:
    def to_pil(self) -> Image.Image: ...

class PdfColorScheme:
    pass

class PdfPage:
    def get_width(self) -> int: ...
    def get_height(self) -> int: ...
    def render(
        self,
        scale: float = ...,
        rotation: int = ...,
        crop: Tuple[float, float, float, float] = ...,
        may_draw_forms: bool = ...,
        bitmap_maker: Callable = ...,
        color_scheme: Union[PdfColorScheme, None] = ...,
        fill_to_stroke: bool = ...,
        **kwargs,
    ) -> PdfBitmap: ...
    def close(self) -> None: ...

class PdfDocument:
    def __init__(self, input, password=None, autoclose=False) -> None: ...
    def __iter__(self) -> Iterator[PdfPage]: ...
    def __getitem__(self, idx: int) -> PdfPage: ...
    def close(self) -> None: ...



import sys
from dataclasses import dataclass
from pathlib import Path

import pytest

import playa
import paves.image as pi

THISDIR = Path(__file__).parent


@pytest.mark.skipif(
    sys.platform.startswith("win") or sys.platform.startswith("darwin"),
    reason="Poppler Probably not Present on Proprietary Platforms",
)
def test_popple() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        images = list(pi.popple(path))
        assert len(images) == 15
        assert all("page_width" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == list(
            range(len(pdf.pages))
        )
        images = list(pi.popple(pdf))
        assert len(images) == len(pdf.pages)
        assert all("page_height" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == list(
            range(len(pdf.pages))
        )
        images = list(pi.popple(pdf.pages[1:6]))
        assert len(images) == 5
        assert all("page_width" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == [1, 2, 3, 4, 5]
        images = list(pi.popple(pdf.pages[[3, 4, 5, 9, 10]]))
        assert len(images) == 5
        assert all("page_height" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == [3, 4, 5, 9, 10]
        images = list(pi.popple(pdf.pages[1]))
        assert len(images) == 1
        assert all("page_width" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == [1]


def test_pdfium() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        images = list(pi.pdfium(path))
        assert len(images) == len(pdf.pages)
        assert all("page_width" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == list(
            range(len(pdf.pages))
        )
        images = list(pi.pdfium(pdf))
        assert len(images) == len(pdf.pages)
        assert all("page_height" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == list(
            range(len(pdf.pages))
        )
        images = list(pi.pdfium(pdf.pages[1:6]))
        assert len(images) == 5
        assert all("page_width" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == [1, 2, 3, 4, 5]
        images = list(pi.pdfium(pdf.pages[[3, 4, 5, 9, 10]]))
        assert len(images) == 5
        assert all("page_height" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == [3, 4, 5, 9, 10]
        images = list(pi.pdfium(pdf.pages[1]))
        assert len(images) == 1
        assert all("page_width" in image.info for image in images)
        assert [image.info["page_index"] for image in images] == [1]


def test_box() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        page = pdf.pages[0]
        img = pi.box(page)
        assert img
        img = pi.box(page, color="red")
        assert img
        img = pi.box(page, color=["green", "orange", "purple"])
        assert img
        img = pi.box(page, dpi=100, color={"text": "red", "image": "green"})
        assert img

        @dataclass
        class PagyThing:
            bbox: playa.Rect
            page: playa.Page

        img = pi.box(PagyThing(bbox=(1, 2, 3, 4), page=page))
        assert img
        img = pi.box([(1, 2, 3, 4), (4, 5, 6, 7)], image=pi.show(page))
        assert img
        img = pi.box(iter([(1, 2, 3, 4), (4, 5, 6, 7)]), image=pi.show(page))
        assert img
        img = pi.box(((1, 2, 3, 4), (4, 5, 6, 7)), image=pi.show(page))
        assert img

        @dataclass
        class BoxyThing:
            bbox: playa.Rect

        img = pi.box(BoxyThing(bbox=(1, 2, 3, 4)), image=pi.show(page))
        assert img
        img = pi.box(
            [BoxyThing(bbox=(1, 2, 3, 4)), BoxyThing(bbox=(4, 5, 6, 7))],
            image=pi.show(page),
        )
        assert img


def test_mark() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        page = pdf.pages[0]
        img = pi.mark(page)
        assert img
        img = pi.mark(page, color="red")
        assert img
        img = pi.mark(page, color=["green", "orange", "purple"])
        assert img
        img = pi.mark(page, dpi=100, color={"text": "red", "image": "green"})
        assert img



"""
Common interface for image converters.
"""

from os import PathLike
from typing import Callable, Iterator, List, Protocol, Tuple, Union
from playa import Document, Page, PageList
from PIL import Image

from paves.exceptions import NotInstalledError


class Converter(Protocol):
    """Protocol for image to PDF converters."""

    def __call__(
        self,
        pdf: Union[str, PathLike, Document, Page, PageList],
        *,
        dpi: int = 0,
        width: int = 0,
        height: int = 0,
    ) -> Iterator[Image.Image]: ...

    __name__: str


CONVERTERS: List[Tuple[int, Converter]] = []


def converter(*, priority: int) -> Callable[[Converter], Converter]:
    """Decorator to register converter functions with priorities."""

    def register(func: Converter) -> Converter:
        CONVERTERS.append((priority, func))
        # We don't care about the inefficiency of this as there are
        # only ever going to be a few of them
        CONVERTERS.sort()
        return func

    return register


def convert(
    pdf: Union[str, PathLike, Document, Page, PageList],
    *,
    dpi: int = 0,
    width: int = 0,
    height: int = 0,
) -> Iterator[Image.Image]:
    """Convert a PDF to images.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.
        dpi: Render to this resolution (default is 72 dpi).
        width: Render to this width in pixels (0 to keep aspect ratio).
        height: Render to this height in pixels (0 to keep aspect ratio).
    Yields:
        Pillow `Image.Image` objects, one per page.  The original page
        width and height in default user space units are available in
        the `info` property of these images as `page_width` and
        `page_height`
    Raises:
        ValueError: Invalid arguments (e.g. both `dpi` and `width`/`height`)
        NotInstalledError: If no renderer is available

    """
    for _, convert in CONVERTERS:
        try:
            for img in convert(pdf, dpi=dpi, width=width, height=height):
                yield img
            break
        except NotInstalledError:
            continue
    else:
        raise NotInstalledError(
            "No converters available, tried: %s"
            % (", ".join(m.__name__ for _, m in CONVERTERS))
        )



"""
PDFMiner.six compatibility, which now just pulls stuff from playa.miner.

This will definitely go away in PAVÉS 1.0.
"""

from playa.miner import (
    ColorSpace,
    GraphicState,
    KWD,
    LIT,
    NameTree,
    NumberTree,
    PDFDocEncoding,
    PDFDocument,
    PDFPage,
    PDFException,
    PDFObjRef,
    PDFObject,
    PSException,
    PSLiteral,
    decode_text,
    pdftypes,
    resolve1,
    resolve_all,
    PathSegment,
    PDFTypeError,
    PDFValueError,
    Plane,
    LAParams,
    LTItem,
    LTText,
    LTComponent,
    LTCurve,
    LTLine,
    LTRect,
    LTImage,
    LTAnno,
    LTChar,
    LTContainer,
    LTExpandableContainer,
    LTTextContainer,
    TextLineElement,
    LTTextLine,
    LTTextLineHorizontal,
    LTTextLineVertical,
    LTTextBox,
    LTTextBoxHorizontal,
    LTTextBoxVertical,
    TextGroupElement,
    LTTextGroup,
    LTTextGroupLRTB,
    LTTextGroupTBRL,
    LTLayoutContainer,
    LTFigure,
    LTPage,
    extract_page,
    extract,
)


__all__ = [
    "ColorSpace",
    "GraphicState",
    "KWD",
    "LIT",
    "NameTree",
    "NumberTree",
    "PDFDocEncoding",
    "PDFDocument",
    "PDFPage",
    "PDFException",
    "PDFObjRef",
    "PDFObject",
    "PSException",
    "PSLiteral",
    "decode_text",
    "extract",
    "extract_page",
    "pdftypes",
    "resolve1",
    "resolve_all",
    "PathSegment",
    "PDFTypeError",
    "PDFValueError",
    "Plane",
    "LAParams",
    "LTItem",
    "LTText",
    "LTComponent",
    "LTCurve",
    "LTLine",
    "LTRect",
    "LTImage",
    "LTAnno",
    "LTChar",
    "LTContainer",
    "LTExpandableContainer",
    "LTTextContainer",
    "TextLineElement",
    "LTTextLine",
    "LTTextLineHorizontal",
    "LTTextLineVertical",
    "LTTextBox",
    "LTTextBoxHorizontal",
    "LTTextBoxVertical",
    "TextGroupElement",
    "LTTextGroup",
    "LTTextGroupLRTB",
    "LTTextGroupTBRL",
    "LTLayoutContainer",
    "LTFigure",
    "LTPage",
]



"""
Convert PDFs to images using pypdfium2.
"""

import contextlib
import functools
from io import BytesIO
from os import PathLike
from typing import (
    TYPE_CHECKING,
    Iterator,
    Tuple,
    Union,
)

from PIL import Image
from playa.document import Document, PageList
from playa.page import Page

from paves.image.converters import converter
from paves.exceptions import NotInstalledError

if TYPE_CHECKING:
    import pypdfium2  # types: ignore


@functools.singledispatch
def _get_pdfium_pages(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Iterator[Tuple[int, "pypdfium2.PdfPage"]]:
    import pypdfium2

    doc = pypdfium2.PdfDocument(pdf)
    for idx, page in enumerate(doc):
        yield idx, page
        page.close()
    doc.close()


@contextlib.contextmanager
def _get_pdfium_doc(pdf: Document) -> Iterator["pypdfium2.PdfDocument"]:
    import pypdfium2

    if pdf._fp is None:
        # Yes, you can actually wrap a BytesIO around an mmap!
        with BytesIO(pdf.buffer) as fp:
            doc = pypdfium2.PdfDocument(fp)
            yield doc
            doc.close()
    else:
        doc = pypdfium2.PdfDocument(pdf._fp)
        yield doc
        doc.close()


@_get_pdfium_pages.register(Document)
def _get_pdfium_pages_doc(pdf: Document) -> Iterator[Tuple[int, "pypdfium2.PdfPage"]]:
    with _get_pdfium_doc(pdf) as doc:
        for idx, page in enumerate(doc):
            yield idx, page
            page.close()


@_get_pdfium_pages.register(Page)
def _get_pdfium_pages_page(page: Page) -> Iterator[Tuple[int, "pypdfium2.PdfPage"]]:
    pdf = page.doc
    assert pdf is not None
    with _get_pdfium_doc(pdf) as doc:
        pdfium_page = doc[page.page_idx]
        yield page.page_idx, pdfium_page
        pdfium_page.close()


@_get_pdfium_pages.register(PageList)
def _get_pdfium_pages_pagelist(
    pages: PageList,
) -> Iterator[Tuple[int, "pypdfium2.PdfPage"]]:
    pdf = pages.doc
    assert pdf is not None
    with _get_pdfium_doc(pdf) as doc:
        for page in pages:
            pdfium_page = doc[page.page_idx]
            yield page.page_idx, pdfium_page
            pdfium_page.close()


@converter(priority=20)
def pdfium(
    pdf: Union[str, PathLike, Document, Page, PageList],
    *,
    dpi: int = 0,
    width: int = 0,
    height: int = 0,
) -> Iterator[Image.Image]:
    """Convert a PDF to images using PyPDFium2

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.
        dpi: Render to this resolution (default is 72 dpi).
        width: Render to this width in pixels.
        height: Render to this height in pixels.
    Yields:
        Pillow `Image.Image` objects, one per page.  Page width and height are
        available in the `info` property of the images.
    Raises:
        ValueError: Invalid arguments (e.g. both `dpi` and `width`/`height`)
        NotInstalledError: If PyPDFium2 is not installed.
    """
    if dpi and (width or height):
        raise ValueError("Cannot specify both `dpi` and `width` or `height`")
    try:
        import pypdfium2  # noqa: F401
    except ImportError as e:
        raise NotInstalledError("PyPDFium2 does not seem to be installed") from e
    for idx, page in _get_pdfium_pages(pdf):
        page_width = page.get_width()
        page_height = page.get_height()
        if width == 0 and height == 0:
            scale = (dpi or 72) / 72
            img = page.render(scale=scale).to_pil()
        else:
            if width and height:
                # Scale to longest side (since pypdfium2 doesn't
                # appear to allow non-1:1 aspect ratio)
                scale = max(width / page_width, height / page_height)
                img = page.render(scale=scale).to_pil()
                # Resize down to desired size
                img = img.resize(size=(width, height))
            elif width:
                scale = width / page.get_width()
                img = page.render(scale=scale).to_pil()
            elif height:
                scale = height / page.get_height()
                img = page.render(scale=scale).to_pil()
        img.info["page_index"] = idx
        img.info["page_width"] = page_width
        img.info["page_height"] = page_height
        yield img



"""
Various ways of converting PDFs to images for feeding them to
models and/or visualisation.`
"""

import functools
import itertools
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Protocol,
    Tuple,
    Union,
    cast,
)

from PIL import Image, ImageDraw, ImageFont
from playa.page import ContentObject, Page, Annotation
from playa.structure import Element
from playa.utils import Rect

from paves.image.converters import convert
from paves.image.poppler import popple
from paves.image.pdfium import pdfium

__all__ = ["convert", "popple", "pdfium", "show", "box", "mark"]


def show(page: Page, dpi: int = 72) -> Image.Image:
    """Show a single page with some reasonable defaults."""
    try:
        return next(convert(page, dpi=dpi))
    except NotImplementedError as e:
        raise ValueError(
            f"Can't call show() on a {type(page).__name__}, "
            "did you mean to call box() or mark()?"
        ) from e


class HasBbox(Protocol):
    bbox: Rect


class HasPage(Protocol):
    page: Page


Boxable = Union[Annotation, ContentObject, Element, HasBbox, Rect]
"""Object for which we can get a bounding box."""
LabelFunc = Callable[[Boxable], Any]
"""Function to get a label for a Boxable."""
BoxFunc = Callable[[Boxable], Union[Rect, None]]
"""Function to get a bounding box for a Boxable."""


@functools.singledispatch
def get_box(obj) -> Union[Rect, None]:
    """Default function to get the bounding box for an object."""
    if hasattr(obj, "bbox"):
        return obj.bbox
    raise RuntimeError(f"Don't know how to get the box for {obj!r}")


@get_box.register(tuple)
def get_box_rect(obj: Rect) -> Union[Rect, None]:
    """Get the bounding box of a bounding box"""
    return obj


@functools.singledispatch
def get_label(obj: Boxable) -> str:
    """Default function to get the label text for an object."""
    return str(obj)


@get_label.register(ContentObject)
def get_label_content(obj: ContentObject) -> str:
    """Get the label text for a ContentObject."""
    return obj.object_type


@get_label.register(Annotation)
def get_label_annotation(obj: Annotation) -> str:
    """Get the default label text for an Annotation.

    Note: This is just a default.
        This is one of many possible options, so you may wish to
        define your own custom LabelFunc.
    """
    return obj.type


@get_label.register(Element)
def get_label_element(obj: Element) -> str:
    """Get the default label text for an Element.

    Note: This is just a default.
        This is one of many possible options, so you may wish to
        define your own custom LabelFunc.
    """
    return obj.type


def _make_boxes(
    obj: Union[
        Annotation,
        ContentObject,
        Element,
        Rect,
        HasBbox,
        Iterable[Union[Boxable, None]],
    ],
) -> Iterable[Union[Boxable, None]]:
    """Put a box into a list of boxes if necessary."""
    # Is it a single Rect? (mypy is incapable of understanding the
    # runtime check here so we need the cast among other things)
    if isinstance(obj, tuple):
        if len(obj) == 4 and all(isinstance(x, (int, float)) for x in obj):
            return [cast(Rect, obj)]
        # This shouldn't be necessary... but mypy needs it
        return list(obj)
    if isinstance(obj, (Annotation, ContentObject, Element)):
        return [obj]
    if hasattr(obj, "bbox"):
        # Ugh, we have to cast
        return [cast(HasBbox, obj)]
    return obj


def _getpage(
    obj: Boxable,
    page: Union[Page, None] = None,
) -> Page:
    if page is None:
        if not hasattr(obj, "page"):
            raise ValueError("Must explicitly specify page or image to show rectangles")
        page = cast(HasPage, obj).page
    if page is None:
        raise ValueError("No page found in object: %r" % (obj,))
    return page


Color = Union[str, Tuple[int, int, int], Tuple[float, float, float]]
"""Type alias for things that can be used as colors."""
Colors = Union[Color, List[Color], Dict[str, Color]]
"""Type alias for colors or collections of colors."""
PillowColor = Union[str, Tuple[int, int, int]]
"""Type alias for things Pillow accepts as colors."""
ColorMaker = Callable[[str], PillowColor]
"""Function that makes a Pillow color for a string label."""
DEFAULT_COLOR_CYCLE: Colors = [
    "blue",
    "orange",
    "green",
    "red",
    "purple",
    "brown",
    "pink",
    "gray",
    "olive",
    "cyan",
]
"""Default color cycle (same as matplotlib)"""


def pillow_color(color: Color) -> PillowColor:
    """Convert colors to a form acceptable to Pillow."""
    if isinstance(color, str):
        return color
    r, g, b = color
    # Would sure be nice if MyPy understood all()
    if isinstance(r, int) and isinstance(g, int) and isinstance(b, int):
        return (r, g, b)
    r, g, b = (int(x * 255) for x in color)
    return (r, g, b)


@functools.singledispatch
def color_maker(spec: Colors, default: Color = "red") -> ColorMaker:
    """Create a function that makes colors."""
    return lambda _: pillow_color(default)


@color_maker.register(str)
@color_maker.register(tuple)
def _color_maker_string(spec: Color, default: Color = "red") -> ColorMaker:
    return lambda _: pillow_color(spec)


@color_maker.register(dict)
def _color_maker_dict(spec: Dict[str, Color], default: Color = "red") -> ColorMaker:
    colors: Dict[str, PillowColor] = {k: pillow_color(v) for k, v in spec.items()}
    pdefault: PillowColor = pillow_color(default)

    def maker(label: str) -> PillowColor:
        return colors.get(label, pdefault)

    return maker


@color_maker.register(list)
def _color_maker_list(spec: List[Color], default: Color = "UNUSED") -> ColorMaker:
    itor = itertools.cycle(spec)
    colors: Dict[str, PillowColor] = {}

    def maker(label: str) -> PillowColor:
        if label not in colors:
            colors[label] = pillow_color(next(itor))
        return colors[label]

    return maker


def box(
    objs: Union[
        Boxable,
        Iterable[Union[Boxable, None]],
    ],
    *,
    color: Colors = DEFAULT_COLOR_CYCLE,
    label: bool = True,
    label_color: Color = "white",
    label_size: float = 9,
    label_margin: float = 1,
    label_fill: bool = True,
    image: Union[Image.Image, None] = None,
    labelfunc: LabelFunc = get_label,
    boxfunc: BoxFunc = get_box,
    dpi: int = 72,
    page: Union[Page, None] = None,
) -> Union[Image.Image, None]:
    """Draw boxes around things in a page of a PDF."""
    draw: ImageDraw.ImageDraw
    scale = dpi / 72
    font = ImageFont.load_default(label_size * scale)
    label_margin *= scale
    make_color = color_maker(color)
    image_page: Union[Page, None] = None
    for obj in _make_boxes(objs):
        if obj is None:
            continue
        if image_page is not None:
            if hasattr(obj, "page"):
                if cast(HasPage, obj).page != image_page:
                    break
        if image is None:
            image_page = _getpage(obj, page)
            image = show(image_page, dpi)
        try:
            box = boxfunc(obj)
            if box is None:  # it has no box
                continue
            left, top, right, bottom = (x * scale for x in box)
        except ValueError:  # it has no content and no box
            continue
        draw = ImageDraw.ImageDraw(image)
        text = str(labelfunc(obj))
        obj_color = make_color(text)
        draw.rectangle((left, top, right, bottom), outline=obj_color)
        if label:
            tl, tt, tr, tb = font.getbbox(text)
            label_box = (
                left,
                top - tb - label_margin * 2,
                left + tr + label_margin * 2,
                top,
            )
            draw.rectangle(
                label_box,
                outline=obj_color,
                fill=obj_color if label_fill else None,
            )
            draw.text(
                xy=(left + label_margin, top - label_margin),
                text=text,
                font=font,
                fill="white" if label_fill else obj_color,
                anchor="ld",
            )
    return image


def mark(
    objs: Union[
        Boxable,
        Iterable[Union[Boxable, None]],
    ],
    *,
    color: Colors = DEFAULT_COLOR_CYCLE,
    transparency: float = 0.75,
    label: bool = False,
    label_color: Color = "white",
    label_size: float = 9,
    label_margin: float = 1,
    outline: bool = False,
    image: Union[Image.Image, None] = None,
    labelfunc: LabelFunc = get_label,
    boxfunc: BoxFunc = get_box,
    dpi: int = 72,
    page: Union[Page, None] = None,
) -> Union[Image.Image, None]:
    """Highlight things in a page of a PDF."""
    overlay: Union[Image.Image, None] = None
    mask: Union[Image.Image, None] = None
    draw: ImageDraw.ImageDraw
    scale = dpi / 72
    font = ImageFont.load_default(label_size * scale)
    alpha = min(255, int(transparency * 255))
    label_margin *= scale
    make_color = color_maker(color)
    image_page: Union[Page, None] = None
    for obj in _make_boxes(objs):
        if obj is None:
            continue
        if image_page is not None:
            if hasattr(obj, "page"):
                if cast(HasPage, obj).page != image_page:
                    break
        if image is None:
            image_page = _getpage(obj, page)
            image = show(image_page, dpi)
        if overlay is None:
            overlay = Image.new("RGB", image.size)
        if mask is None:
            mask = Image.new("L", image.size, 255)
        try:
            box = boxfunc(obj)
            if box is None:  # it has no box
                continue
            left, top, right, bottom = (x * scale for x in box)
        except ValueError:  # it has no content and no box
            continue
        draw = ImageDraw.ImageDraw(overlay)
        text = str(labelfunc(obj))
        obj_color = make_color(text)
        draw.rectangle((left, top, right, bottom), fill=obj_color)
        mask_draw = ImageDraw.ImageDraw(mask)
        mask_draw.rectangle((left, top, right, bottom), fill=alpha)
        if outline:
            draw.rectangle((left, top, right, bottom), outline="black")
            mask_draw.rectangle((left, top, right, bottom), outline=0)
        if label:
            tl, tt, tr, tb = font.getbbox(text)
            label_box = (
                left,
                top - tb - label_margin * 2,
                left + tr + label_margin * 2,
                top,
            )
            draw.rectangle(
                label_box,
                outline=obj_color,
                fill=obj_color,
            )
            mask_draw.rectangle(
                label_box,
                fill=alpha,
            )
            if outline:
                draw.rectangle(
                    label_box,
                    outline="black",
                )
                mask_draw.rectangle(
                    label_box,
                    outline=0,
                )
                draw.text(
                    xy=(left + label_margin, top - label_margin),
                    text=text,
                    font=font,
                    fill="black",
                    anchor="ld",
                )
                mask_draw.text(
                    xy=(left + label_margin, top - label_margin),
                    text=text,
                    font=font,
                    fill=0,
                    anchor="ld",
                )
            else:
                draw.text(
                    xy=(left + label_margin, top - label_margin),
                    text=text,
                    font=font,
                    fill="white",
                    anchor="ld",
                )
    if image is None:
        return None
    if overlay is not None and mask is not None:
        return Image.composite(image, overlay, mask)
    else:
        return image



"""
Detect tables using RT-DETR models from IBM Docling project, or
Microsoft Table Transformer (which is also DETR).
"""

import logging
from functools import singledispatch
from os import PathLike
from typing import Any, Iterator, List, Tuple, Union

import paves.image as pi
import playa
from paves.tables.detectors import detector
from paves.tables.table import TableObject
from playa import Document, Page, PageList, Rect

LOGGER = logging.getLogger(__name__)


@singledispatch
def _get_pages(pdf: Union[str, PathLike, Document, Page, PageList]) -> Iterator[Page]:
    raise NotImplementedError


@_get_pages.register(str)
@_get_pages.register(PathLike)
def _get_pages_path(pdf: Union[str, PathLike]) -> Iterator[Page]:
    with playa.open(pdf) as doc:
        yield from doc.pages


@_get_pages.register
def _get_pages_pagelist(pagelist: PageList) -> Iterator[Page]:
    yield from pagelist


@_get_pages.register
def _get_pages_doc(doc: Document) -> Iterator[Page]:
    yield from doc.pages


@_get_pages.register
def _get_pages_page(page: Page) -> Iterator[Page]:
    yield page


def make_rect(box: List[Union[int, float]]) -> Rect:
    """Verify and create a bounding box as tuple of ints."""
    rect = tuple(round(x) for x in box)
    if len(rect) != 4:
        raise TypeError(f"Rectangle does not have 4 corners: {box}")
    return rect


def detect_objects(
    pdf: Union[str, PathLike, Document, Page, PageList],
    model_name: str,
    *,
    model_kwargs: dict[str, Any] | None = None,
    threshold: float = 0.5,
) -> Iterator[Tuple[int, List[Tuple[str, Rect]]]]:
    """Iterate over all text objects in a PDF, page, or pages"""
    import torch
    from transformers import AutoImageProcessor, AutoModelForObjectDetection

    processor = AutoImageProcessor.from_pretrained(model_name, use_fast=True)
    # FIXME: sorry, AMD owners, and everybody else, this will get fixed
    device = "cuda" if torch.cuda.is_available() else "cpu"
    torch_device = torch.device(device)
    if model_kwargs is None:
        model_kwargs = {}
    model = AutoModelForObjectDetection.from_pretrained(model_name, **model_kwargs).to(
        torch_device
    )

    # Concoct some arguments for pi.convert (FIXME: should be able to
    # pass it processor.size directly)
    dpi = width = height = 0
    if "width" in processor.size and "height" in processor.size:
        width = processor.size["width"]
        height = processor.size["height"]
    elif "longest_edge" in processor.size:
        # FIXME: This isn't really what it means, but it works anyway
        width = height = processor.size["longest_edge"]
    else:
        # Render it big and let the model figure it out
        dpi = 144

    # We could do this in a batch, but that easily runs out of memory
    with torch.inference_mode():
        for image in pi.convert(pdf, dpi=dpi, width=width, height=height):
            inputs = processor(images=[image], return_tensors="pt").to(torch_device)
            outputs = model(**inputs)
            results = processor.post_process_object_detection(
                outputs,
                target_sizes=[(image.info["page_height"], image.info["page_width"])],
                threshold=threshold,
            )
            boxes: List[Tuple[str, Rect]] = []
            for label, box in zip(results[0]["labels"], results[0]["boxes"]):
                name = model.config.id2label[label.item()]
                bbox = make_rect(box.tolist())
                boxes.append((name, bbox))
            yield image.info["page_index"], boxes


@detector(priority=10)
def docling_heron(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Union[Iterator[TableObject], None]:
    """Identify tables in a PDF or one of its pages using Docling Project
    layout model.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.

    Returns:
      An iterator over `TableObject`, or `None`, if the model can't be used

    """
    try:
        detected = detect_objects(pdf, "docling-project/docling-layout-heron")
    except ImportError:
        return None

    def itor() -> Iterator[TableObject]:
        for page, (page_idx, objects) in zip(_get_pages(pdf), detected):
            assert page.page_idx == page_idx
            for label, bbox in objects:
                if label == "Table":
                    yield TableObject.from_bbox(page, bbox)

    return itor()


@detector(priority=20)
def table_transformer(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Union[Iterator[TableObject], None]:
    """Identify tables in a PDF or one of its pages using Microsoft Table
    Transformer model.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.

    Returns:
      An iterator over `TableObject`, or `None`, if the model can't be used

    """
    try:
        detected = detect_objects(
            pdf,
            "microsoft/table-transformer-detection",
            model_kwargs={"revision": "no_timm"},
            threshold=0.9,
        )
    except ImportError:
        return None

    def itor() -> Iterator[TableObject]:
        for page, (page_idx, objects) in zip(_get_pages(pdf), detected):
            assert page.page_idx == page_idx
            for label, bbox in objects:
                yield TableObject.from_bbox(page, bbox)

    return itor()



"""Benchmark table detection with logical structure."""

import time
from pathlib import Path

import playa
import paves.tables as pt


def benchmark(detector, path: Path):
    with playa.open(path) as doc:
        for table in detector(doc):
            print(table.page.page_idx, table.bbox)


def benchmark_pagelist(detector, path: Path):
    with playa.open(path) as doc:
        for table in detector(doc.pages):
            print(table.page.page_idx, table.bbox)


def benchmark_pages(detector, path: Path):
    with playa.open(path) as doc:
        for page in doc.pages:
            for table in detector(page):
                print(page.page_idx, table.bbox)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("pdf", type=Path)
    parser.add_argument("--over", choices=["doc", "page", "pagelist"], default="doc")
    parser.add_argument(
        "--detector",
        choices=["structure", "docling_heron", "table_transformer"],
        default="structure",
    )
    args = parser.parse_args()

    detector = pt.detector(args.detector)
    if detector is None:
        parser.error(f"Unknown detector {args.detector}")
    if args.over == "doc":
        start = time.time()
        benchmark(detector, args.pdf)
        multi_time = time.time() - start
        print("Full document took %.2fs" % multi_time)
    elif args.over == "pagelist":
        start = time.time()
        benchmark_pagelist(detector, args.pdf)
        multi_time = time.time() - start
        print("PageList took %.2fs" % multi_time)
    elif args.over == "page":
        start = time.time()
        benchmark_pages(detector, args.pdf)
        multi_time = time.time() - start
        print("Page took %.2fs" % multi_time)



"""
Table types.
"""

from copy import copy
from typing import Iterable, Iterator, Tuple, Union

from playa import Page
from playa.content import GraphicState, MarkedContent
from playa.page import Annotation
from playa.pdftypes import Matrix, Rect
from playa.structure import (
    Element,
    ContentItem,
    ContentObject as StructContentObject,
)
from playa.worker import _ref_page
from dataclasses import dataclass

from playa.content import ContentObject
from playa.utils import get_bound_rects


@dataclass
class TableObject(ContentObject):
    """Table on one page of a PDF.

    This **is** a ContentObject and can be treated as one (notably
    with `paves.image` functions).

    It could either come from a logical structure element, or it could
    simply be a bounding box (as detected by some sort of visual
    model).  While these `TableObject`s will never span multiple
    pages, the underlying logical structure element may do so.

    If there is underlying logical structure it is accessible (as with
    all other `ContentObject` instances) through the `parent`
    property.  This is currently the only way to detect multi-page
    tables through this interface (they will have an equivalent
    `parent`).

    Note that the graphics state and coordinate transformation matrix
    may just be the page defaults, if Machine Learning™ was used to
    detect the table in a rendered image of the page.

    """

    _bbox: Union[Rect, None]
    _parent: Union[Element, None]

    @property
    def bbox(self) -> Union[Rect, None]:
        # _bbox takes priority as we *could* have both
        if self._bbox is not None:
            return self._bbox
        elif self._parent is not None:
            # Try to get it from the element but only if it has the
            # same page as us (otherwise it will be wrong!)
            if self._parent.page is self.page:
                bbox = self._parent.bbox
                if bbox is not None:
                    return bbox
            # We always have a page even if self._parent doesn't
            return get_bound_rects(
                item.bbox
                for item in self._parent.contents
                if item.page is self.page and item.bbox is not None
            )
        else:
            # This however should never happen
            return None

    @classmethod
    def from_bbox(cls, page: Page, bbox: Rect) -> "TableObject":
        # Use default values
        return cls(
            _pageref=_ref_page(page),
            _parentkey=None,
            gstate=GraphicState(),
            ctm=page.ctm,
            mcstack=(),
            _bbox=bbox,
            _parent=None,
        )

    @classmethod
    def from_element(
        cls,
        el: Element,
        page: Page,
        contents: Union[Iterable[Union[ContentItem, StructContentObject]], None] = None,
    ) -> Union["TableObject", None]:
        if contents is None:
            contents = el.contents
        # Find a ContentObject so we can get a bbox, mcstack, ctm
        # (they might not be *correct* of course, but oh well)
        gstate: Union[GraphicState, None] = None
        ctm: Union[Matrix, None] = None
        mcstack: Union[Tuple[MarkedContent, ...], None] = None
        bbox: Union[Rect, None] = None
        for kid in contents:
            # For multi-page tables, skip any contents on a different page
            if kid.page != page:
                continue
            if isinstance(kid, StructContentObject):
                obj = kid.obj
                if obj is None:
                    continue
                elif isinstance(obj, Annotation):
                    # FIXME: for the moment just ignore these
                    continue
                else:
                    gstate = copy(obj.gstate)
                    ctm = obj.ctm
                    mcstack = obj.mcstack
                    bbox = obj.bbox
                    break
            elif isinstance(kid, ContentItem):
                # It's a ContentItem
                try:
                    cobj = next(iter(kid))
                except StopIteration:
                    continue
                gstate = copy(cobj.gstate)
                ctm = cobj.ctm
                mcstack = cobj.mcstack
                break
        else:
            # No contents, no table for you!
            return None
        return cls(
            _pageref=_ref_page(page),
            _parentkey=None,
            gstate=gstate,
            ctm=ctm,
            mcstack=mcstack,
            _bbox=bbox,
            _parent=el,
        )

    def __iter__(self) -> Iterator["RowObject"]:
        if self._parent is not None:
            for el in self._parent:
                if isinstance(el, Element):
                    row = RowObject.from_element(el)
                    if row is not None:
                        yield row


class RowObject(ContentObject):
    """Row in a table.

    This contains table cells and can be iterated over accordingly.
    """

    @classmethod
    def from_element(
        cls,
        el: Element,
    ) -> Union["RowObject", None]:
        return None

    def __iter__(self) -> Iterator["CellObject"]:
        if self._parent is not None:
            for el in self._parent:
                if isinstance(el, Element):
                    row = CellObject.from_element(el)
                    if row is not None:
                        yield row


class CellObject(ContentObject):
    """Cell in a table.

    This might contain structure (including other tables) if it came
    from PDF logical structure, but otherwise it probably doesn't.
    You can access this through the `parent` attribute if so.  We will
    also try to transparently extract text from it for you.

    In some cases (not always) we are able to determine if this cell
    spans multiple rows or columns.

    """

    @classmethod
    def from_element(
        cls,
        el: Element,
    ) -> Union["CellObject", None]:
        return None



from itertools import groupby
from pathlib import Path

import playa
import paves.text as px

THISDIR = Path(__file__).parent


def test_words_pdf() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        list(px.words(pdf))


def test_words_page() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        texts = [w.chars for w in px.words(pdf.pages[0])]
        assert texts == [
            "Réserve",
            "de",
            "biodiversité",
            "projetée",
            "de",
            "la",
            "Station-de-",
            "Biologie-des-",
            "Laurentides",
            "Février",
            "2009",
        ]


def test_words_pagelist() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        list(px.words(pdf.pages[0:4]))


def test_words_path() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    list(px.words(path))


def test_lines() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        lines = [
            " ".join(w.chars for w in words)
            for line, words in groupby(px.words(pdf.pages[0]), px.line)
        ]
        assert lines == [
            "Réserve de",
            "biodiversité",
            "projetée de la",
            "Station-de-",
            "Biologie-des-",
            "Laurentides",
            "Février 2009",
        ]



"""
Convert PDFs to images using poppler command-line tools.
"""

import functools
import shutil
import subprocess
import tempfile
from os import PathLike
from pathlib import Path
from typing import (
    Iterator,
    List,
    Tuple,
    Union,
)

import playa
from PIL import Image
from playa.document import Document, PageList
from playa.page import Page

from paves.exceptions import NotInstalledError
from paves.image.converters import converter

PATH_TO_POPPLER: Union[str, None] = shutil.which("pdftoppm")


def make_poppler_args(dpi: int, width: int, height: int) -> List[str]:
    args = []
    if width or height:
        args.extend(
            [
                "-scale-to-x",
                str(width or -1),  # -1 means use aspect ratio
                "-scale-to-y",
                str(height or -1),
            ]
        )
    if not args:
        args.extend(["-r", str(dpi or 72)])
    return args


@functools.singledispatch
def _popple(pdf, tempdir: Path, args: List[str]) -> List[Tuple[int, float, float]]:
    raise NotImplementedError


@_popple.register(str)
@_popple.register(PathLike)
def _popple_path(
    pdf: Union[str, PathLike], tempdir: Path, args: List[str]
) -> List[Tuple[int, float, float]]:
    subprocess.run(
        [
            "pdftoppm",
            *args,
            str(pdf),
            tempdir / "ppm",
        ],
        check=True,
    )
    with playa.open(pdf) as doc:
        return [(page.page_idx, page.width, page.height) for page in doc.pages]


@_popple.register(Document)
def _popple_doc(
    pdf: Document, tempdir: Path, args: List[str]
) -> List[Tuple[int, float, float]]:
    pdfpdf = tempdir / "pdf.pdf"
    # FIXME: This is... not great (can we popple in a pipeline please?)
    with open(pdfpdf, "wb") as outfh:
        outfh.write(pdf.buffer)
    subprocess.run(
        [
            "pdftoppm",
            *args,
            str(pdfpdf),
            tempdir / "ppm",
        ],
        check=True,
    )
    pdfpdf.unlink()
    return [(page.page_idx, page.width, page.height) for page in pdf.pages]


@_popple.register(Page)
def _popple_page(
    pdf: Page, tempdir: Path, args: List[str]
) -> List[Tuple[int, float, float]]:
    assert pdf.doc is not None  # bug in PLAYA-PDF, oops, it cannot be None
    pdfpdf = tempdir / "pdf.pdf"
    with open(pdfpdf, "wb") as outfh:
        outfh.write(pdf.doc.buffer)
    page_number = pdf.page_idx + 1
    subprocess.run(
        [
            "pdftoppm",
            *args,
            "-f",
            str(page_number),
            "-l",
            str(page_number),
            str(pdfpdf),
            tempdir / "ppm",
        ],
        check=True,
    )
    pdfpdf.unlink()
    return [(pdf.page_idx, pdf.width, pdf.height)]


@_popple.register(PageList)
def _popple_pages(
    pdf: PageList, tempdir: Path, args: List[str]
) -> List[Tuple[int, float, float]]:
    pdfpdf = tempdir / "pdf.pdf"
    assert pdf[0].doc is not None  # bug in PLAYA-PDF, oops, it cannot be None
    with open(pdfpdf, "wb") as outfh:
        outfh.write(pdf[0].doc.buffer)
    pages = sorted(page.page_idx + 1 for page in pdf)
    itor = iter(pages)
    first = last = next(itor)
    spans = []
    while True:
        try:
            next_last = next(itor)
        except StopIteration:
            spans.append((first, last))
            break
        if next_last > last + 1:
            spans.append((first, last))
            first = last = next_last
        else:
            last = next_last
    assert PATH_TO_POPPLER is not None
    for first, last in spans:
        subprocess.run(
            [
                PATH_TO_POPPLER,
                *args,
                "-f",
                str(first),
                "-l",
                str(last),
                str(pdfpdf),
                tempdir / "ppm",
            ],
            check=True,
        )
    pdfpdf.unlink()
    return [(page.page_idx, page.width, page.height) for page in pdf]


@converter(priority=10)
def popple(
    pdf: Union[str, PathLike, Document, Page, PageList],
    *,
    dpi: int = 0,
    width: int = 0,
    height: int = 0,
) -> Iterator[Image.Image]:
    """Convert a PDF to images using Poppler's pdftoppm.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.
        dpi: Render to this resolution (default is 72 dpi).
        width: Render to this width in pixels.
        height: Render to this height in pixels.
    Yields:
        Pillow `Image.Image` objects, one per page.
    Raises:
        ValueError: Invalid arguments (e.g. both `dpi` and `width`/`height`)
        NotInstalledError: If Poppler is not installed.
    """
    if dpi and (width or height):
        raise ValueError("Cannot specify both `dpi` and `width` or `height`")
    try:
        subprocess.run(["pdftoppm", "-h"], capture_output=True)
    except FileNotFoundError as e:
        raise NotInstalledError("Poppler does not seem to be installed") from e
    args = make_poppler_args(dpi, width, height)
    with tempfile.TemporaryDirectory() as tempdir:
        temppath = Path(tempdir)
        # FIXME: Possible to Popple in a Parallel Pipeline
        page_sizes = _popple(pdf, temppath, args)
        for (page_idx, page_width, page_height), ppm in zip(
            page_sizes,
            (path for path in sorted(temppath.iterdir()) if path.suffix == ".ppm"),
        ):
            img = Image.open(ppm)
            img.info["page_index"] = page_idx
            img.info["page_width"] = page_width
            img.info["page_height"] = page_height
            yield img



"""
Table detection using PDF logical structure.
"""

from functools import singledispatch
from itertools import groupby
from typing import Iterable, Iterator, Union
from operator import attrgetter
from os import PathLike

import playa
from playa import Document, Page, PageList
from playa.structure import (
    Element,
)

from paves.tables.detectors import detector
from paves.tables.table import TableObject


@singledispatch
def table_elements(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Iterator[Element]:
    """Iterate over all text objects in a PDF, page, or pages"""
    raise NotImplementedError(f"Not implemented for {type(pdf)}")


@table_elements.register(str)
@table_elements.register(PathLike)
def table_elements_path(pdf: Union[str, PathLike]) -> Iterator[Element]:
    with playa.open(pdf) as doc:
        # NOTE: This *must* be `yield from` or else we will return a
        # useless iterator (as the document will go out of scope)
        yield from table_elements_doc(doc)


@table_elements.register
def table_elements_doc(pdf: Document) -> Iterator[Element]:
    structure = pdf.structure
    if structure is None:
        raise TypeError("Document has no logical structure")
    return structure.find_all("Table")


@table_elements.register
def table_elements_pagelist(pages: PageList) -> Iterator[Element]:
    if pages.doc.structure is None:
        raise TypeError("Document has no logical structure")
    for page in pages:
        yield from table_elements_page(page)


@table_elements.register
def table_elements_page(page: Page) -> Iterator[Element]:
    # page.structure can actually never be None (why?)
    if page.structure is None:
        raise TypeError("Page has no ParentTree")
    if len(page.structure) == 0:
        raise TypeError("Page has no marked content")
    return page.structure.find_all("Table")


def table_elements_to_objects(
    elements: Iterable[Element], page: Union[Page, None] = None
) -> Iterator[TableObject]:
    """Make TableObjects from Elements."""
    for el in elements:
        # It usually has a page, but it can also span multiple pages
        # if this is the case.  So a page passed explicitly here
        # should take precedence.
        for kidpage, kids in groupby(el.contents, attrgetter("page")):
            if kidpage is None:
                continue
            if page is not None and kidpage is not page:
                continue
            table = TableObject.from_element(el, kidpage, kids)
            if table is not None:
                yield table


@detector(priority=0)
def structure(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Union[Iterator[TableObject], None]:
    """Identify tables in a PDF or one of its pages using logical structure.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.

    Returns:
      An iterator over `TableObject`, or `None`, if there is no
      logical structure (this will cause a TypeError, if you don't
      check for it).
    """
    page = pdf if isinstance(pdf, Page) else None
    try:
        return table_elements_to_objects(table_elements(pdf), page)
    except TypeError:  # means that structure is None
        return None



from pathlib import Path

import playa
import paves.tables as pb

THISDIR = Path(__file__).parent


def test_tables() -> None:
    path = THISDIR / "contrib" / "Rgl-1314-2021-Z-en-vigueur-20240823.pdf"
    with playa.open(path) as pdf:
        tables = pb.tables(pdf.pages[300])
        assert tables is not None
        assert len(list(tables)) == 2


def test_no_tables() -> None:
    path = THISDIR / "contrib" / "PSC_Station.pdf"
    with playa.open(path) as pdf:
        tables = pb.tables(pdf)
        assert tables is not None
        assert len(list(tables)) == 0


def test_multi_page_tables() -> None:
    path = THISDIR / "data" / "multi-page-table.pdf"
    with playa.open(path) as pdf:
        itor = pb.tables_orelse(pdf)
        assert itor is not None
        tables = list(itor)
        assert len(tables) == 2
        assert tables[0].parent == tables[1].parent
        table1 = next(pb.tables(pdf.pages[0]))
        table2 = next(pb.tables(pdf.pages[1]))
        assert table1.page == pdf.pages[0]
        assert table2.page == pdf.pages[1]
        assert table1.bbox != table2.bbox



"""
Simple and not at all Java-damaged interface for table detection.
"""

from paves.tables.detectors import lookup as detector
from paves.tables.detectors import tables, tables_orelse
from paves.tables.detr import docling_heron as tables_detr
from paves.tables.detr import table_transformer as tables_tatr
from paves.tables.structure import structure as tables_structure

__all__ = [
    "tables",
    "tables_orelse",
    "detector",
    "tables_structure",
    "tables_tatr",
    "tables_detr",
]



"""
Various somewhat-more-heuristic ways of guessing, getting, and
processing text in PDFs.
"""

import operator
from dataclasses import dataclass
from functools import singledispatch
from os import PathLike
from typing import Iterator, List, Union, cast

import playa
from playa.content import ContentObject, GlyphObject, TextBase, TextObject
from playa.document import Document, PageList
from playa.page import Page
from playa.pdftypes import Point, Matrix

# For convenience in grouping word/text/glyph objects
line = operator.attrgetter("line")
font = operator.attrgetter("font")
fontname = operator.attrgetter("fontname")
fontbase = operator.attrgetter("fontbase")
size = operator.attrgetter("size")
textfont = operator.attrgetter("textfont")
chars = operator.attrgetter("chars")


@dataclass
class WordObject(TextBase):
    """
    "Word" in a PDF.

    This is heuristically determined, either by explicit whitespace
    (if you're lucky enough to have a Tagged PDF) or by a sufficient
    gap between adjacent glyphs (otherwise).

    It otherwise behaves just like a `TextObject`.  You can iterate
    over its glyphs, etc.  But, as a treat, these glyphs are
    "finalized" so you don't have to worry about inconsistent graphics
    states and so forth, and you also get some convenience properties.

    The origin of the curent (logical) line is also available, to
    facilitate grouping words into lines, if you so desire (simply
    use `itertools.groupby(words, paves.text.line)`)
    """

    _glyphs: List[GlyphObject]
    _next_origin: Point
    line: Point

    def __iter__(self) -> Iterator["ContentObject"]:
        return iter(self._glyphs)

    @property
    def matrix(self) -> Matrix:
        return self._glyphs[0].matrix

    @property
    def chars(self) -> str:
        return "".join(g.text for g in self._glyphs if g.text is not None)

    @property
    def origin(self) -> Point:
        return self._glyphs[0].origin

    @property
    def displacement(self) -> Point:
        ax, ay = self.origin
        bx, by = self._next_origin
        return bx - ax, by - ay


def word_break(
    glyph: GlyphObject, predicted_origin: Point, prev_displacement: Point
) -> bool:
    """Heuristically predict a word break based on the predicted origin
    from the previous glyph."""
    if glyph.text == " ":
        return True
    x, y = glyph.origin
    px, py = predicted_origin
    if glyph.font.vertical:
        glyph_offset = y - py
        _, displacement = prev_displacement
        if glyph.page.space == "screen":
            glyph_offset = -glyph_offset
            displacement = -displacement
    else:
        glyph_offset = x - px
        displacement, _ = prev_displacement
    # If there's a space, *or* if we are before the prev glyph
    return glyph_offset > 0.5 or glyph_offset < -displacement


def line_break(glyph: GlyphObject, predicted_origin: Point) -> bool:
    """Heuristically predict a line break based on the predicted origin
    from the previous glyph."""
    x, y = glyph.origin
    px, py = predicted_origin
    if glyph.font.vertical:
        line_offset = x - px
    else:
        line_offset = y - py
        if glyph.page.space == "screen":
            line_offset = -line_offset
    return line_offset < 0 or line_offset > 100  # FIXME: arbitrary!


@singledispatch
def text_objects(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Iterator[TextObject]:
    """Iterate over all text objects in a PDF, page, or pages"""
    raise NotImplementedError


@text_objects.register(str)
@text_objects.register(PathLike)
def text_objects_path(pdf: Union[str, PathLike]) -> Iterator[TextObject]:
    with playa.open(pdf) as doc:
        # NOTE: This *must* be `yield from` or else we will return a
        # useless iterator (as the document will go out of scope)
        yield from text_objects_doc(doc)


@text_objects.register
def text_objects_doc(pdf: Document) -> Iterator[TextObject]:
    return text_objects_pagelist(pdf.pages)


@text_objects.register
def text_objects_pagelist(pagelist: PageList) -> Iterator[TextObject]:
    for page in pagelist:
        yield from text_objects_page(page)


@text_objects.register
def text_objects_page(page: Page) -> Iterator[TextObject]:
    return page.texts


def _add_point(a: Point, b: Point) -> Point:
    return a[0] + b[0], a[1] + b[1]


def words(
    pdf: Union[str, PathLike, Document, Page, PageList],
) -> Iterator[WordObject]:
    """Extract "words" (i.e. whitespace-separated text cells) from a
    PDF or one of its pages.

    Args:
        pdf: PLAYA-PDF document, page, pages, or path to a PDF.

    Yields:
        `WordObject` objects, which can be visualized with `paves.image`
        functions, or you can do various other things with them too.
    """
    glyphs: List[GlyphObject] = []
    predicted_origin: Union[None, Point] = None
    prev_disp: Union[None, Point] = None
    line_origin: Union[None, Point] = None
    for obj in text_objects(pdf):
        for glyph in obj:
            if line_origin is None:
                line_origin = glyph.origin
            if predicted_origin and prev_disp:
                new_word = word_break(glyph, predicted_origin, prev_disp)
                new_line = line_break(glyph, predicted_origin)
                if glyphs and (new_word or new_line):
                    yield WordObject(
                        _pageref=glyphs[0]._pageref,
                        _parentkey=glyphs[0]._parentkey,
                        gstate=glyphs[0].gstate,  # Not necessarily correct!
                        ctm=glyphs[0].ctm,  # Not necessarily correct!
                        mcstack=glyphs[0].mcstack,  # Not necessarily correct!
                        _glyphs=glyphs,
                        _next_origin=predicted_origin,
                        line=line_origin,
                    )
                    glyphs = []
                if new_line:
                    line_origin = glyph.origin
            if glyph.text is not None and glyph.text != " ":
                glyphs.append(cast(GlyphObject, glyph.finalize()))
            prev_disp = glyph.displacement
            predicted_origin = _add_point(glyph.origin, prev_disp)
    if predicted_origin and line_origin and glyphs:
        yield WordObject(
            _pageref=glyphs[0]._pageref,
            _parentkey=glyphs[0]._parentkey,
            gstate=glyphs[0].gstate,  # Not necessarily correct!
            ctm=glyphs[0].ctm,  # Not necessarily correct!
            mcstack=glyphs[0].mcstack,  # Not necessarily correct!
            _glyphs=glyphs,
            _next_origin=predicted_origin,
            line=line_origin,
        )
